﻿/*
 * WCF Self hosting sample from www.WayneCliffordBarker.co.za
*/

using ActiveSTS.Handlers;
using System;
using System.Collections.Generic;
using System.Configuration;
using System.IdentityModel;
using System.IdentityModel.Configuration;
using System.IdentityModel.Tokens;
using System.Linq;
using System.Security;
using System.Security.Claims;
using System.Security.Cryptography.X509Certificates;
using System.ServiceModel.Security;
using System.Text;
using System.Threading.Tasks;

namespace STS
{
    public class TokenService : SecurityTokenService
    {
        public TokenService(SecurityTokenServiceConfiguration securityTokenServiceConfiguration)
            : base(securityTokenServiceConfiguration)
        {
    }

        public static X509Certificate2 GetCertificate(string subjectName)
        {
            var store = new X509Store(StoreName.My, StoreLocation.LocalMachine);
            X509Certificate2Collection certificates = null;
            store.Open(OpenFlags.ReadOnly);

            try
            {
                certificates = store.Certificates;
                var certs = certificates.OfType<X509Certificate2>().Where(x => x.SubjectName.Name.Equals(subjectName, StringComparison.OrdinalIgnoreCase)).ToList();

                if (certs.Count == 0)
                    throw new SecurityException(string.Format("No certificate for subject {0}", subjectName));
                else if (certs.Count > 1)
                    throw new SecurityException(string.Format("There are multiple certificates for {0}", subjectName));

                return new X509Certificate2(certs[0]);
            }
            finally
            {
                if (certificates != null)
                {
                    for (var i = 0; i < certificates.Count; i++)
                        certificates[i].Reset();
                }

                store.Close();
            }
        }

        protected override System.Security.Claims.ClaimsIdentity GetOutputClaimsIdentity(System.Security.Claims.ClaimsPrincipal principal, System.IdentityModel.Protocols.WSTrust.RequestSecurityToken request, Scope scope)
        {
            if (principal == null)
            {
                throw new ArgumentNullException("principal", "'principal' cannot be null.");
            }

            if (request == null)
            {
                throw new ArgumentNullException("request", "'request' cannot be null.");
            }

            if (scope == null)
            {
                throw new ArgumentNullException("scope", "'scope' cannot be null.");
            }

            // Retrieve the authenticated Identity
            ClaimsIdentity authenticatedIdentity = principal.Identity as ClaimsIdentity;
            if (authenticatedIdentity == null)
            {
                throw new ArgumentNullException("authenticatedIdentity", "'authenticatedIdentity' cannot be null.");
            }

            return authenticatedIdentity;
        }

        protected override Scope GetScope(System.Security.Claims.ClaimsPrincipal principal, System.IdentityModel.Protocols.WSTrust.RequestSecurityToken request)
        {
            // Need to validate request AppliesTo url against a list of known allowed Passive STS sites.
            if (principal == null)
            {
                throw new ArgumentNullException("principal", "'principal' cannot be null.");
            }

            if (request == null)
            {
                throw new ArgumentNullException("request", "'request' cannot be null.");
            }

            
            var scope = new Scope(request.AppliesTo.Uri.OriginalString, SecurityTokenServiceConfiguration.SigningCredentials);
            string certificateName = ConfigurationManager.AppSettings["CertificateSubject"];
            var encryptingCertificate = GetCertificate(certificateName);
            var encryptingCredentials = new X509EncryptingCredentials(encryptingCertificate);
            scope.EncryptingCredentials = encryptingCredentials;
            // Enable encryption (true by default)
            scope.TokenEncryptionRequired = true;
            scope.SymmetricKeyEncryptionRequired = true;
            scope.ReplyToAddress = request.ReplyTo;

            return scope;
        }

        public static SecurityTokenServiceConfiguration GetSTSConfiguration()
        {
            string certificateIssuer = ConfigurationManager.AppSettings["CertificateIssuer"];
            string IssuerUri = ConfigurationManager.AppSettings["IssuerUri"];

            var cert = GetCertificate(ConfigurationManager.AppSettings["CertificateSubject"]);
            var signingCredentials = new X509SigningCredentials(cert);
            var config = new SecurityTokenServiceConfiguration(certificateIssuer, signingCredentials);

            config.TokenIssuerName = signingCredentials.Certificate.Issuer;
            config.SecurityTokenHandlers.AddOrReplace(new CustomAuthenticationHandler());
            config.SecurityTokenService = typeof(TokenService);
            config.CertificateValidationMode = X509CertificateValidationMode.None;

            ConfigurationBasedIssuerNameRegistry issuerReg = new ConfigurationBasedIssuerNameRegistry();
            issuerReg.AddTrustedIssuer(cert.Thumbprint.ToLowerInvariant(), IssuerUri);
            config.IssuerNameRegistry = issuerReg;

            // Specifies the certificate use to decrypt the federation token. 
            var sessionTransforms = new List<CookieTransform>(new CookieTransform[]
            {
                    new DeflateCookieTransform(),
                    new RsaEncryptionCookieTransform(cert),
                    new RsaSignatureCookieTransform(cert)
            });
            var sessionHandler = new SessionSecurityTokenHandler(sessionTransforms.AsReadOnly());
            config.SecurityTokenHandlers.AddOrReplace(sessionHandler);

            return config;
        }
    }         
}
