﻿/*
 * WCF Self hosting sample from www.WayneCliffordBarker.co.za
*/

using System;
using System.Collections.Generic;
using System.Configuration;
using System.IdentityModel.Protocols.WSTrust;
using System.IdentityModel.Selectors;
using System.IdentityModel.Tokens;
using System.IO;
using System.Linq;
using System.Security;
using System.Security.Claims;
using System.Security.Cryptography.X509Certificates;
using System.ServiceModel;
using System.ServiceModel.Security;
using System.Threading;
using System.Xml;

namespace ActiveSTSProxy
{
    public static class Proxy
    {
        private static string audienceUri;
        private static string signingCertificateName;

        public static Boolean SignIn(string userName, string password, out SessionSecurityToken sessionToken)
        {
            sessionToken = null;
            try
            {
                ValidateConfiguration();
                GenericXmlSecurityToken genericToken = RequestTrustToken(userName, password, audienceUri);

                if (genericToken != null)
                {
                    ClaimsPrincipal tokenClaimsPrincipal = ParseToken(genericToken);
                    if (tokenClaimsPrincipal != null)
                    {
                        // Generate a session token
                        sessionToken = new SessionSecurityToken(tokenClaimsPrincipal);

                        // Assigns the principal from the token to the client thread.
                        Thread.CurrentPrincipal = tokenClaimsPrincipal;
                        return true;
                    }
                }

                return false;
            }
            catch (MessageSecurityException e)
            {
                return false;
            }
        }

        public static void SignOut(ref SessionSecurityToken sessionToken)
        {
            if (sessionToken == null)
            {
                throw new ArgumentNullException("sessionToken");
            }

            if (sessionToken.ClaimsPrincipal == null)
            {
                throw new ArgumentException("sessionToken.ClaimsPrincipal");
            }

            if (sessionToken.ClaimsPrincipal.Identity == null)
            {
                throw new ArgumentException("sessionToken.ClaimsPrincipal.Identity");
            }

            if (sessionToken.ClaimsPrincipal.Identity.IsAuthenticated)
            {
                ClaimsIdentity sessionIdentity = sessionToken.ClaimsPrincipal.Identity as ClaimsIdentity;

                while ((sessionToken.ClaimsPrincipal.Identity as ClaimsIdentity).Claims.Count() > 0)
                {
                    Claim sessionClaim = sessionToken.ClaimsPrincipal.Claims.First();
                    sessionIdentity.RemoveClaim(sessionClaim);
                }

                sessionToken = null;
            }
        }

        private static void ValidateConfiguration()
        {
            audienceUri = ConfigurationManager.AppSettings["TokenAudienceUri"];
            if (string.IsNullOrEmpty(audienceUri))
            {
                throw new ArgumentNullException("TokenAudienceUri", "No 'AudienceUri' defined in configuration.");
            }

            signingCertificateName = ConfigurationManager.AppSettings["CertificateSubject"];
            if (string.IsNullOrEmpty(signingCertificateName))
            {
                throw new ArgumentNullException("SigningCertificateName", "No 'SigningCertificateName' defined in configuration.");
            }
        }

        private static WSTrustChannel CreateWSTrustChannel(string userName, string password)
        {
            WS2007HttpBinding binding = new WS2007HttpBinding(SecurityMode.Message);
            binding.Security.Message.ClientCredentialType = MessageCredentialType.UserName;
            binding.Security.Message.NegotiateServiceCredential = true;
            binding.Security.Message.EstablishSecurityContext = false;

            TimeSpan timeSpan = new TimeSpan(0, 5, 0);
            binding.SendTimeout = timeSpan;
            binding.ReceiveTimeout = timeSpan;
            binding.OpenTimeout = timeSpan;
            binding.CloseTimeout = timeSpan;

            EndpointIdentity identity = EndpointIdentity.CreateDnsIdentity(ConfigurationManager.AppSettings["CertificateDnsIdentity"]);
            EndpointAddress address = new EndpointAddress(new Uri(ConfigurationManager.AppSettings["ActiveSTSUrl"]), identity);

            WSTrustChannelFactory factory = new WSTrustChannelFactory(binding, address);

            factory.TrustVersion = TrustVersion.WSTrust13;            
            factory.Credentials.ServiceCertificate.Authentication.CertificateValidationMode = System.ServiceModel.Security.X509CertificateValidationMode.None;            
            factory.Credentials.ServiceCertificate.Authentication.RevocationMode = System.Security.Cryptography.X509Certificates.X509RevocationMode.NoCheck;
            factory.Credentials.SupportInteractive = false;
            factory.Credentials.UserName.UserName = userName;
            factory.Credentials.UserName.Password = password;

            return (WSTrustChannel)factory.CreateChannel();
        }

        private static GenericXmlSecurityToken RequestTrustToken(string userName, string password, string appliesToUrl)
        {
            WSTrustChannel channel = null;
            try
            {
                channel = CreateWSTrustChannel(userName, password);

                RequestSecurityToken request = new RequestSecurityToken(System.IdentityModel.Protocols.WSTrust.RequestTypes.Issue)
                {
                    AppliesTo = new EndpointReference(appliesToUrl)
                };

                RequestSecurityTokenResponse response = null;
                var genericToken = channel.Issue(request, out response);
                return genericToken as System.IdentityModel.Tokens.GenericXmlSecurityToken;
            }
            catch
            {
                if (channel != null && channel.State == CommunicationState.Faulted)
                {
                    channel.Abort();
                }

                throw;
            }
            finally
            {
                if (channel != null && channel.State != CommunicationState.Faulted)
                {
                    channel.Close();
                }
            }
        }

        public static X509Certificate2 GetCertificate(string subjectName)
        {
            var store = new X509Store(StoreName.My, StoreLocation.LocalMachine);
            X509Certificate2Collection certificates = null;
            store.Open(OpenFlags.ReadOnly);

            try
            {
                certificates = store.Certificates;
                var certs = certificates.OfType<X509Certificate2>().Where(x => x.SubjectName.Name.Equals(subjectName, StringComparison.OrdinalIgnoreCase)).ToList();

                if (certs.Count == 0)
                    throw new SecurityException(string.Format("No certificate for subject {0}", subjectName));
                else if (certs.Count > 1)
                    throw new SecurityException(string.Format("There are multiple certificates for {0}", subjectName));

                return new X509Certificate2(certs[0]);
            }
            finally
            {
                if (certificates != null)
                {
                    for (var i = 0; i < certificates.Count; i++)
                        certificates[i].Reset();
                }

                store.Close();
            }
        }

        private static ClaimsPrincipal ParseToken(GenericXmlSecurityToken token)
        {
            using (var stringReader = new StringReader(token.TokenXml.OuterXml))
            {
                using (var samlReader = XmlReader.Create(stringReader))
                {
                    var tokenHandlers = SecurityTokenHandlerCollection.CreateDefaultSecurityTokenHandlerCollection();

                    SecurityTokenHandlerConfiguration config = tokenHandlers.Configuration;
                    var securityTokens = new List<SecurityToken>()
                    {
                    new X509SecurityToken(GetCertificate(signingCertificateName))
                    };

                    config.ServiceTokenResolver = SecurityTokenResolver.CreateDefaultSecurityTokenResolver(securityTokens.AsReadOnly(), false);
                    config.CertificateValidator = X509CertificateValidator.None;
                    

                    config.IssuerTokenResolver = new X509CertificateStoreTokenResolver(StoreName.My, StoreLocation.LocalMachine);
                    config.IssuerNameRegistry = new TrustedIssuerNameRegistry();


                    // This is a list of clients that are allowed to read the token.
                    config.AudienceRestriction.AllowedAudienceUris.Add(new Uri(audienceUri));
                    SecurityToken samlToken = tokenHandlers.ReadToken(samlReader);

                    var tokenIdentity = tokenHandlers.ValidateToken(samlToken).FirstOrDefault();
                    return new ClaimsPrincipal(tokenIdentity);
                }
            }
        }
    }
}